import ctypes
from typing import Tuple

import numpy as np

from ._c_lib import _get_library


class SphericalHarmonics:
    """
    Spherical harmonics calculator, which computes the real spherical harmonics
    :math:`Y^m_l` up to degree ``l_max``. The calculated spherical harmonics
    are consistent with the definition of real spherical harmonics from Wikipedia.

    The ``SphericalHarmonics`` object computes prefactors and initializes buffers
    upon creation

    >>> import numpy as np
    >>> import sphericart as sc
    >>> sh = sc.SphericalHarmonics(l_max=8)

    Then, the :py:func:`compute` method can be called on an array of 3D
    Cartesian points to compute the spherical harmonics

    >>> xyz = np.random.normal(size=(10, 3))
    >>> sh_values = sh.compute(xyz)
    >>> sh_values.shape
    (10, 81)

    In order to also compute derivatives, you can use

    >>> sh_values, sh_grads = sh.compute_with_gradients(xyz)
    >>> sh_grads.shape
    (10, 3, 81)

    which returns the gradient as a tensor with size
    ``(n_samples, 3, (l_max+1)**2)``.

    :param l_max: the maximum degree of the spherical harmonics to be calculated

    :return: a calculator, in the form of a ``SphericalHarmonics`` object
    """

    def __init__(self, l_max: int):
        self._l_max = l_max

        self._lib = _get_library()

        # intialize both a double precision and a single-precision calculator.
        # we will decide which one to use depending on the dtype of the data
        self._calculator = self._lib.sphericart_spherical_harmonics_new(l_max)
        self._calculator_f = self._lib.sphericart_spherical_harmonics_new_f(l_max)

        # this allows to check the number of threads that are used
        # it is 1 if there is no OpenMP available
        self._omp_num_threads = (
            self._lib.sphericart_spherical_harmonics_omp_num_threads(self._calculator)
        )

    def __del__(self):
        if self._calculator is not None:
            self._lib.sphericart_spherical_harmonics_delete(self._calculator)
            self._calculator = None

        if self._calculator_f is not None:
            self._lib.sphericart_spherical_harmonics_delete_f(self._calculator_f)
            self._calculator_f = None

    def compute(self, xyz: np.ndarray) -> np.ndarray:
        """
        Calculates the spherical harmonics for a set of 3D points, whose
        coordinates are given by the ``xyz`` array.

        >>> import numpy as np
        >>> import sphericart as sc
        >>> sh = sc.SphericalHarmonics(l_max=8)
        >>> xyz = np.random.normal(size=(10, 3))
        >>> sh_values = sh.compute(xyz)
        >>> sh_values.shape
        (10, 81)

        :param xyz:
            The Cartesian coordinates of the 3D points, as an array with
            shape ``(n_samples, 3)``

        :return:
            An array of shape ``(n_samples, (l_max+1)**2)`` containing all the
            spherical harmonics up to degree `l_max` in lexicographic order.
            For example, if ``l_max = 2``, The last axis will correspond to
            spherical harmonics with ``(l, m) = (0, 0), (1, -1), (1, 0), (1,
            1), (2, -2), (2, -1), (2, 0), (2, 1), (2, 2)``, in this order.
        """
        if self._calculator is None or self._calculator_f is None:
            raise ValueError("can not use a deleted calculator")

        if not isinstance(xyz, np.ndarray):
            raise TypeError("xyz must be a numpy array")

        if not (xyz.dtype == np.float32 or xyz.dtype == np.float64):
            raise TypeError("xyz must be a numpy array of 32 or 64-bit floats")

        if len(xyz.shape) != 2 or xyz.shape[1] != 3:
            raise ValueError("xyz array must be a `N x 3` array")

        # make xyz contiguous before taking a pointer to it
        xyz = np.ascontiguousarray(xyz)

        n_samples = xyz.shape[0]
        xyz_length = n_samples * 3

        sph = np.empty((n_samples, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        sph_length = n_samples * (self._l_max + 1) ** 2

        if xyz.dtype == np.float64:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

            self._lib.sphericart_spherical_harmonics_compute_array(
                self._calculator, xyz_ptr, xyz_length, sph_ptr, sph_length
            )
        elif xyz.dtype == np.float32:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

            self._lib.sphericart_spherical_harmonics_compute_array_f(
                self._calculator_f, xyz_ptr, xyz_length, sph_ptr, sph_length
            )

        return sph

    def compute_with_gradients(self, xyz: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculates the spherical harmonics for a set of 3D points, whose
        coordinates are in the ``xyz`` array, together with their Cartesian
        derivatives.

        >>> import numpy as np
        >>> import sphericart as sc
        >>> sh = sc.SphericalHarmonics(l_max=8)
        >>> xyz = np.random.normal(size=(10, 3))
        >>> sh_values, sh_grads = sh.compute_with_gradients(xyz)
        >>> sh_grads.shape
        (10, 3, 81)

        :param xyz: The Cartesian coordinates of the 3D points, as an array with
            shape ``(n_samples, 3)``.

        :return: A tuple containing:

            - an array of shape ``(n_samples, (l_max+1)**2)`` containing all the
              spherical harmonics up to degree ``l_max`` in lexicographic order.
              For example, if ``l_max = 2``, The last axis will correspond to
              spherical harmonics with ``(l, m) = (0, 0), (1, -1), (1, 0), (1,
              1), (2, -2), (2, -1), (2, 0), (2, 1), (2, 2)``, in this order.
            - an array of shape ``(n_samples, 3, (l_max+1)**2)`` containing all
              the spherical harmonics' derivatives up to degree ``l_max``. The
              last axis is organized in the same way as in the spherical
              harmonics return array, while the second-to-last axis refers to
              derivatives in the x, y, and z directions, respectively.

        """
        if self._calculator is None or self._calculator_f is None:
            raise ValueError("can not use a deleted calculator")

        if not isinstance(xyz, np.ndarray):
            raise TypeError("xyz must be a numpy array")

        if not (xyz.dtype == np.float32 or xyz.dtype == np.float64):
            raise TypeError("xyz must be a numpy array of 32 or 64-bit floats")

        if len(xyz.shape) != 2 or xyz.shape[1] != 3:
            raise ValueError("xyz array must be a `N x 3` array")

        # make xyz contiguous before taking a pointer to it
        xyz = np.ascontiguousarray(xyz)

        n_samples = xyz.shape[0]
        xyz_length = n_samples * 3

        sph = np.empty((n_samples, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        sph_length = n_samples * (self._l_max + 1) ** 2
        dsph = np.empty((n_samples, 3, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        dsph_length = n_samples * 3 * (self._l_max + 1) ** 2

        if xyz.dtype == np.float64:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            dsph_ptr = dsph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

            self._lib.sphericart_spherical_harmonics_compute_array_with_gradients(
                self._calculator,
                xyz_ptr,
                xyz_length,
                sph_ptr,
                sph_length,
                dsph_ptr,
                dsph_length,
            )
        elif xyz.dtype == np.float32:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            dsph_ptr = dsph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

            self._lib.sphericart_spherical_harmonics_compute_array_with_gradients_f(
                self._calculator_f,
                xyz_ptr,
                xyz_length,
                sph_ptr,
                sph_length,
                dsph_ptr,
                dsph_length,
            )

        return sph, dsph

    def compute_with_hessians(
        self, xyz: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculates the spherical harmonics for a set of 3D points, whose
        coordinates are in the ``xyz`` array, together with their Cartesian
        derivatives and second derivatives.

        >>> import numpy as np
        >>> import sphericart as sc
        >>> sh = sc.SphericalHarmonics(l_max=8, normalized=False)
        >>> xyz = np.random.normal(size=(10, 3))
        >>> sh_values, sh_grads, sh_hessians = sh.compute_with_hessians(xyz)
        >>> sh_values.shape
        (10, 81)
        >>> sh_grads.shape
        (10, 3, 81)
        >>> sh_hessians.shape
        (10, 3, 3, 81)

        :param xyz: The Cartesian coordinates of the 3D points, as an array with
            shape ``(n_samples, 3)``.

        :return: A tuple containing:

            - an array of shape ``(n_samples, (l_max+1)**2)`` containing all the
              spherical harmonics up to degree ``l_max`` in lexicographic order.
              For example, if ``l_max = 2``, The last axis will correspond to
              spherical harmonics with ``(l, m) = (0, 0), (1, -1), (1, 0), (1,
              1), (2, -2), (2, -1), (2, 0), (2, 1), (2, 2)``, in this order.
            - an array of shape ``(n_samples, 3, (l_max+1)**2)`` containing all
              the spherical harmonics' derivatives up to degree ``l_max``. The
              last axis is organized in the same way as in the spherical
              harmonics return array, while the second-to-last axis refers to
              derivatives in the the x, y, and z directions, respectively.
            - an array of shape ``(n_samples, 3, 3, (l_max+1)**2)`` containing all
              the spherical harmonics' second derivatives up to degree ``l_max``.
              The last axis is organized in the same way as in the spherical
              harmonics return array, while the two intermediate axes represent the
              Hessian dimensions.

        """
        if self._calculator is None or self._calculator_f is None:
            raise ValueError("can not use a deleted calculator")

        if not isinstance(xyz, np.ndarray):
            raise TypeError("xyz must be a numpy array")

        if not (xyz.dtype == np.float32 or xyz.dtype == np.float64):
            raise TypeError("xyz must be a numpy array of 32 or 64-bit floats")

        if len(xyz.shape) != 2 or xyz.shape[1] != 3:
            raise ValueError("xyz array must be a `N x 3` array")

        # make xyz contiguous before taking a pointer to it
        xyz = np.ascontiguousarray(xyz)

        n_samples = xyz.shape[0]
        xyz_length = n_samples * 3

        sph = np.empty((n_samples, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        sph_length = n_samples * (self._l_max + 1) ** 2
        dsph = np.empty((n_samples, 3, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        dsph_length = n_samples * 3 * (self._l_max + 1) ** 2
        ddsph = np.empty((n_samples, 3, 3, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        ddsph_length = n_samples * 9 * (self._l_max + 1) ** 2

        if xyz.dtype == np.float64:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            dsph_ptr = dsph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            ddsph_ptr = ddsph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

            self._lib.sphericart_spherical_harmonics_compute_array_with_hessians(
                self._calculator,
                xyz_ptr,
                xyz_length,
                sph_ptr,
                sph_length,
                dsph_ptr,
                dsph_length,
                ddsph_ptr,
                ddsph_length,
            )
        elif xyz.dtype == np.float32:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            dsph_ptr = dsph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            ddsph_ptr = ddsph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

            self._lib.sphericart_spherical_harmonics_compute_array_with_gradients_f(
                self._calculator_f,
                xyz_ptr,
                xyz_length,
                sph_ptr,
                sph_length,
                dsph_ptr,
                dsph_length,
                ddsph_ptr,
                ddsph_length,
            )

        return sph, dsph, ddsph


class SolidHarmonics:
    """
    Solid harmonics calculator, up to degree ``l_max``.

    This class computes the solid harmonics, a non-normalized form of the real
    spherical harmonics, i.e. :math:`r^l Y^m_l`. These scaled spherical harmonics
    are polynomials in the Cartesian coordinates of the input points, and they
    are therefore faster to compute.

    :param l_max: the maximum degree of the solid harmonics to be calculated

    :return: a calculator, in the form of a ``SolidHarmonics`` object
    """

    def __init__(self, l_max: int):
        self._l_max = l_max

        self._lib = _get_library()

        # intialize both a double precision and a single-precision calculator.
        # we will decide which one to use depending on the dtype of the data
        self._calculator = self._lib.sphericart_solid_harmonics_new(l_max)
        self._calculator_f = self._lib.sphericart_solid_harmonics_new_f(l_max)

        # this allows to check the number of threads that are used
        # it is 1 if there is no OpenMP available
        self._omp_num_threads = self._lib.sphericart_solid_harmonics_omp_num_threads(
            self._calculator
        )

    def __del__(self):
        if self._calculator is not None:
            self._lib.sphericart_solid_harmonics_delete(self._calculator)
            self._calculator = None

        if self._calculator_f is not None:
            self._lib.sphericart_solid_harmonics_delete_f(self._calculator_f)
            self._calculator_f = None

    def compute(self, xyz: np.ndarray) -> np.ndarray:
        """
        Same as ``SphericalHarmonics.compute``, but for the solid harmonics.
        """
        if self._calculator is None or self._calculator_f is None:
            raise ValueError("can not use a deleted calculator")

        if not isinstance(xyz, np.ndarray):
            raise TypeError("xyz must be a numpy array")

        if not (xyz.dtype == np.float32 or xyz.dtype == np.float64):
            raise TypeError("xyz must be a numpy array of 32 or 64-bit floats")

        if len(xyz.shape) != 2 or xyz.shape[1] != 3:
            raise ValueError("xyz array must be a `N x 3` array")

        # make xyz contiguous before taking a pointer to it
        xyz = np.ascontiguousarray(xyz)

        n_samples = xyz.shape[0]
        xyz_length = n_samples * 3

        sph = np.empty((n_samples, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        sph_length = n_samples * (self._l_max + 1) ** 2

        if xyz.dtype == np.float64:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

            self._lib.sphericart_solid_harmonics_compute_array(
                self._calculator, xyz_ptr, xyz_length, sph_ptr, sph_length
            )
        elif xyz.dtype == np.float32:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

            self._lib.sphericart_solid_harmonics_compute_array_f(
                self._calculator_f, xyz_ptr, xyz_length, sph_ptr, sph_length
            )

        return sph

    def compute_with_gradients(self, xyz: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Same as ``SphericalHarmonics.compute_with_gradients``, but for the solid
        harmonics.
        """
        if self._calculator is None or self._calculator_f is None:
            raise ValueError("can not use a deleted calculator")

        if not isinstance(xyz, np.ndarray):
            raise TypeError("xyz must be a numpy array")

        if not (xyz.dtype == np.float32 or xyz.dtype == np.float64):
            raise TypeError("xyz must be a numpy array of 32 or 64-bit floats")

        if len(xyz.shape) != 2 or xyz.shape[1] != 3:
            raise ValueError("xyz array must be a `N x 3` array")

        # make xyz contiguous before taking a pointer to it
        xyz = np.ascontiguousarray(xyz)

        n_samples = xyz.shape[0]
        xyz_length = n_samples * 3

        sph = np.empty((n_samples, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        sph_length = n_samples * (self._l_max + 1) ** 2
        dsph = np.empty((n_samples, 3, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        dsph_length = n_samples * 3 * (self._l_max + 1) ** 2

        if xyz.dtype == np.float64:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            dsph_ptr = dsph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

            self._lib.sphericart_solid_harmonics_compute_array_with_gradients(
                self._calculator,
                xyz_ptr,
                xyz_length,
                sph_ptr,
                sph_length,
                dsph_ptr,
                dsph_length,
            )
        elif xyz.dtype == np.float32:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            dsph_ptr = dsph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

            self._lib.sphericart_solid_harmonics_compute_array_with_gradients_f(
                self._calculator_f,
                xyz_ptr,
                xyz_length,
                sph_ptr,
                sph_length,
                dsph_ptr,
                dsph_length,
            )

        return sph, dsph

    def compute_with_hessians(
        self, xyz: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Same as ``SphericalHarmonics.compute_with_hessians``, but for the solid
        harmonics.
        """
        if self._calculator is None or self._calculator_f is None:
            raise ValueError("can not use a deleted calculator")

        if not isinstance(xyz, np.ndarray):
            raise TypeError("xyz must be a numpy array")

        if not (xyz.dtype == np.float32 or xyz.dtype == np.float64):
            raise TypeError("xyz must be a numpy array of 32 or 64-bit floats")

        if len(xyz.shape) != 2 or xyz.shape[1] != 3:
            raise ValueError("xyz array must be a `N x 3` array")

        # make xyz contiguous before taking a pointer to it
        xyz = np.ascontiguousarray(xyz)

        n_samples = xyz.shape[0]
        xyz_length = n_samples * 3

        sph = np.empty((n_samples, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        sph_length = n_samples * (self._l_max + 1) ** 2
        dsph = np.empty((n_samples, 3, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        dsph_length = n_samples * 3 * (self._l_max + 1) ** 2
        ddsph = np.empty((n_samples, 3, 3, (self._l_max + 1) ** 2), dtype=xyz.dtype)
        ddsph_length = n_samples * 9 * (self._l_max + 1) ** 2

        if xyz.dtype == np.float64:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            dsph_ptr = dsph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
            ddsph_ptr = ddsph.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

            self._lib.sphericart_solid_harmonics_compute_array_with_hessians(
                self._calculator,
                xyz_ptr,
                xyz_length,
                sph_ptr,
                sph_length,
                dsph_ptr,
                dsph_length,
                ddsph_ptr,
                ddsph_length,
            )
        elif xyz.dtype == np.float32:
            xyz_ptr = xyz.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            sph_ptr = sph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            dsph_ptr = dsph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))
            ddsph_ptr = ddsph.ctypes.data_as(ctypes.POINTER(ctypes.c_float))

            self._lib.sphericart_solid_harmonics_compute_array_with_gradients_f(
                self._calculator_f,
                xyz_ptr,
                xyz_length,
                sph_ptr,
                sph_length,
                dsph_ptr,
                dsph_length,
                ddsph_ptr,
                ddsph_length,
            )

        return sph, dsph, ddsph



import ctypes
import os
import sys


_HERE = os.path.realpath(os.path.dirname(__file__))


class sphericart_spherical_harmonics_calculator_t(ctypes.c_void_p):
    pass


class sphericart_spherical_harmonics_calculator_f_t(ctypes.c_void_p):
    pass


class sphericart_solid_harmonics_calculator_t(ctypes.c_void_p):
    pass


class sphericart_solid_harmonics_calculator_f_t(ctypes.c_void_p):
    pass


def setup_functions(lib):
    lib.sphericart_spherical_harmonics_new.restype = (
        sphericart_spherical_harmonics_calculator_t
    )
    lib.sphericart_spherical_harmonics_new.argtypes = [
        ctypes.c_size_t,
    ]

    lib.sphericart_spherical_harmonics_new_f.restype = (
        sphericart_spherical_harmonics_calculator_f_t
    )
    lib.sphericart_spherical_harmonics_new_f.argtypes = [
        ctypes.c_size_t,
    ]

    lib.sphericart_spherical_harmonics_delete.restype = None
    lib.sphericart_spherical_harmonics_delete.argtypes = [
        sphericart_spherical_harmonics_calculator_t
    ]

    lib.sphericart_spherical_harmonics_delete_f.restype = None
    lib.sphericart_spherical_harmonics_delete_f.argtypes = [
        sphericart_spherical_harmonics_calculator_f_t
    ]

    lib.sphericart_spherical_harmonics_compute_array.restype = None
    lib.sphericart_spherical_harmonics_compute_array.argtypes = [
        sphericart_spherical_harmonics_calculator_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
    ]

    lib.sphericart_spherical_harmonics_compute_array_with_gradients.restype = None
    lib.sphericart_spherical_harmonics_compute_array_with_gradients.argtypes = [
        sphericart_spherical_harmonics_calculator_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
    ]

    lib.sphericart_spherical_harmonics_compute_array_with_hessians.restype = None
    lib.sphericart_spherical_harmonics_compute_array_with_hessians.argtypes = [
        sphericart_spherical_harmonics_calculator_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
    ]

    lib.sphericart_spherical_harmonics_compute_array_f.restype = None
    lib.sphericart_spherical_harmonics_compute_array_f.argtypes = [
        sphericart_spherical_harmonics_calculator_f_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
    ]

    lib.sphericart_spherical_harmonics_compute_array_with_gradients_f.restype = None
    lib.sphericart_spherical_harmonics_compute_array_with_gradients_f.argtypes = [
        sphericart_spherical_harmonics_calculator_f_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
    ]

    lib.sphericart_spherical_harmonics_compute_array_with_hessians_f.restype = None
    lib.sphericart_spherical_harmonics_compute_array_with_hessians_f.argtypes = [
        sphericart_spherical_harmonics_calculator_f_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
    ]

    lib.sphericart_spherical_harmonics_omp_num_threads.restype = int
    lib.sphericart_spherical_harmonics_omp_num_threads.argtypes = [
        sphericart_spherical_harmonics_calculator_t,
    ]

    lib.sphericart_spherical_harmonics_omp_num_threads_f.restype = int
    lib.sphericart_spherical_harmonics_omp_num_threads_f.argtypes = [
        sphericart_spherical_harmonics_calculator_f_t,
    ]

    lib.sphericart_solid_harmonics_new.restype = sphericart_solid_harmonics_calculator_t
    lib.sphericart_solid_harmonics_new.argtypes = [
        ctypes.c_size_t,
    ]

    lib.sphericart_solid_harmonics_new_f.restype = (
        sphericart_solid_harmonics_calculator_f_t
    )
    lib.sphericart_solid_harmonics_new_f.argtypes = [
        ctypes.c_size_t,
    ]

    lib.sphericart_solid_harmonics_delete.restype = None
    lib.sphericart_solid_harmonics_delete.argtypes = [
        sphericart_solid_harmonics_calculator_t
    ]

    lib.sphericart_solid_harmonics_delete_f.restype = None
    lib.sphericart_solid_harmonics_delete_f.argtypes = [
        sphericart_solid_harmonics_calculator_f_t
    ]

    lib.sphericart_solid_harmonics_compute_array.restype = None
    lib.sphericart_solid_harmonics_compute_array.argtypes = [
        sphericart_solid_harmonics_calculator_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
    ]

    lib.sphericart_solid_harmonics_compute_array_with_gradients.restype = None
    lib.sphericart_solid_harmonics_compute_array_with_gradients.argtypes = [
        sphericart_solid_harmonics_calculator_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
    ]

    lib.sphericart_solid_harmonics_compute_array_with_hessians.restype = None
    lib.sphericart_solid_harmonics_compute_array_with_hessians.argtypes = [
        sphericart_solid_harmonics_calculator_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_double),
        ctypes.c_size_t,
    ]

    lib.sphericart_solid_harmonics_compute_array_f.restype = None
    lib.sphericart_solid_harmonics_compute_array_f.argtypes = [
        sphericart_solid_harmonics_calculator_f_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
    ]

    lib.sphericart_solid_harmonics_compute_array_with_gradients_f.restype = None
    lib.sphericart_solid_harmonics_compute_array_with_gradients_f.argtypes = [
        sphericart_solid_harmonics_calculator_f_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
    ]

    lib.sphericart_solid_harmonics_compute_array_with_hessians_f.restype = None
    lib.sphericart_solid_harmonics_compute_array_with_hessians_f.argtypes = [
        sphericart_solid_harmonics_calculator_f_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
        ctypes.POINTER(ctypes.c_float),
        ctypes.c_size_t,
    ]

    lib.sphericart_solid_harmonics_omp_num_threads.restype = int
    lib.sphericart_solid_harmonics_omp_num_threads.argtypes = [
        sphericart_solid_harmonics_calculator_t,
    ]

    lib.sphericart_solid_harmonics_omp_num_threads_f.restype = int
    lib.sphericart_solid_harmonics_omp_num_threads_f.argtypes = [
        sphericart_solid_harmonics_calculator_f_t,
    ]


class LibraryFinder(object):
    def __init__(self):
        self._cached_dll = None

    def __call__(self):
        if self._cached_dll is None:
            path = _lib_path()
            self._cached_dll = ctypes.cdll.LoadLibrary(path)
            setup_functions(self._cached_dll)

        return self._cached_dll


def _lib_path():
    if sys.platform.startswith("darwin"):
        windows = False
        name = "libsphericart.dylib"
    elif sys.platform.startswith("linux"):
        windows = False
        name = "libsphericart.so"
    elif sys.platform.startswith("win"):
        windows = True
        name = "sphericart.dll"
    else:
        raise ImportError("Unknown platform. Please edit this file")

    path = os.path.join(os.path.join(_HERE, "lib"), name)

    if os.path.isfile(path):
        if windows:
            _check_dll(path)
        return path

    raise ImportError("Could not find sphericart shared library at " + path)


def _check_dll(path):
    """
    Check if the DLL pointer size matches Python (32-bit or 64-bit)
    """
    import platform
    import struct

    IMAGE_FILE_MACHINE_I386 = 332
    IMAGE_FILE_MACHINE_AMD64 = 34404

    machine = None
    with open(path, "rb") as fd:
        header = fd.read(2).decode(encoding="utf-8", errors="strict")
        if header != "MZ":
            raise ImportError(path + " is not a DLL")
        else:
            fd.seek(60)
            header = fd.read(4)
            header_offset = struct.unpack("<L", header)[0]
            fd.seek(header_offset + 4)
            header = fd.read(2)
            machine = struct.unpack("<H", header)[0]

    arch = platform.architecture()[0]
    if arch == "32bit":
        if machine != IMAGE_FILE_MACHINE_I386:
            raise ImportError("Python is 32-bit, but this DLL is not")
    elif arch == "64bit":
        if machine != IMAGE_FILE_MACHINE_AMD64:
            raise ImportError("Python is 64-bit, but this DLL is not")
    else:
        raise ImportError("Could not determine pointer size of Python")


_get_library = LibraryFinder()



import os
import subprocess
import sys
import uuid

from setuptools import Extension, setup
from setuptools.command.bdist_egg import bdist_egg
from setuptools.command.build_ext import build_ext
from wheel.bdist_wheel import bdist_wheel


ROOT = os.path.realpath(os.path.dirname(__file__))
SPHERICART_ARCH_NATIVE = os.environ.get("SPHERICART_ARCH_NATIVE", "ON")


class universal_wheel(bdist_wheel):
    # When building the wheel, the `wheel` package assumes that if we have a
    # binary extension then we are linking to `libpython.so`; and thus the wheel
    # is only usable with a single python version. This is not the case for
    # here, and the wheel will be compatible with any Python >=3.7. This is
    # tracked in https://github.com/pypa/wheel/issues/185, but until then we
    # manually override the wheel tag.
    def get_tag(self):
        tag = bdist_wheel.get_tag(self)
        # tag[2:] contains the os/arch tags, we want to keep them
        return ("py3", "none") + tag[2:]


class cmake_ext(build_ext):
    """Build the native library using cmake"""

    def run(self):
        source_dir = os.path.join(ROOT, "sphericart")
        build_dir = os.path.join(ROOT, "build", "cmake-build")
        install_dir = os.path.join(os.path.realpath(self.build_lib), "sphericart")

        os.makedirs(build_dir, exist_ok=True)

        cmake_options = [
            "-DCMAKE_BUILD_TYPE=Release",
            f"-DCMAKE_INSTALL_PREFIX={install_dir}",
            "-DBUILD_SHARED_LIBS=ON",
            f"-DSPHERICART_ARCH_NATIVE={SPHERICART_ARCH_NATIVE}",
            # Do not create multiple versionned files (libsphericart.so.0.5.0,
            # libsphericart.so.0.5, libsphericart.so), but only the main one
            "-DCMAKE_PLATFORM_NO_VERSIONED_SONAME=ON",
        ]

        CUDA_HOME = os.environ.get("CUDA_HOME")

        if CUDA_HOME is not None:
            cmake_options.append(f"-DCUDA_TOOLKIT_ROOT_DIR={CUDA_HOME}")
            cmake_options.append("-DSPHERICART_ENABLE_CUDA=ON")

        if sys.platform.startswith("darwin"):
            cmake_options.append("-DCMAKE_OSX_DEPLOYMENT_TARGET:STRING=11.0")

        # ARCHFLAGS is used by cibuildwheel to pass the requested arch to the
        # compilers
        ARCHFLAGS = os.environ.get("ARCHFLAGS")
        if ARCHFLAGS is not None:
            cmake_options.append(f"-DCMAKE_C_FLAGS={ARCHFLAGS}")
            cmake_options.append(f"-DCMAKE_CXX_FLAGS={ARCHFLAGS}")

        subprocess.run(
            ["cmake", source_dir, *cmake_options],
            cwd=build_dir,
            check=True,
        )

        build_command = [
            "cmake",
            "--build",
            build_dir,
            "--parallel",
            "2",  # only two jobs to avoid OOM, we don't have many files
            "--target",
            "install",
        ]

        subprocess.run(build_command, check=True)


class bdist_egg_disabled(bdist_egg):
    """Disabled version of bdist_egg

    Prevents setup.py install performing setuptools' default easy_install,
    which it should never ever do.
    """

    def run(self):
        sys.exit(
            "Aborting implicit building of eggs. "
            + "Use `pip install .` or `python setup.py bdist_wheel && pip "
            + "install dist/sphericart-*.whl` to install from source."
        )


if __name__ == "__main__":
    extras_require = {}
    SPHERICART_TORCH = os.path.realpath(os.path.join(ROOT, "sphericart-torch"))
    SPHERICART_JAX = os.path.realpath(os.path.join(ROOT, "sphericart-jax"))

    # when creating a sdist for release, we should never use local dependencies
    SPHERICART_NO_LOCAL_DEPS = os.environ.get("SPHERICART_NO_LOCAL_DEPS", "0") == "1"

    if not SPHERICART_NO_LOCAL_DEPS and os.path.exists(SPHERICART_TORCH):
        # we are building from a checkout

        # add a random uuid to the file url to prevent pip from using a cached
        # wheel for sphericart-torch, and force it to re-build from scratch
        uuid_ = uuid.uuid4()
        extras_require["torch"] = (
            f"sphericart-torch @ file://{SPHERICART_TORCH}?{uuid_}"
        )
    else:
        # installing wheel/sdist
        extras_require["torch"] = "sphericart-torch"

    if not SPHERICART_NO_LOCAL_DEPS and os.path.exists(SPHERICART_JAX):
        # we are building from a checkout

        # add a random uuid to the file url to prevent pip from using a cached
        # wheel for sphericart-jax, and force it to re-build from scratch
        uuid_ = uuid.uuid4()
        extras_require["jax"] = f"sphericart-jax @ file://{SPHERICART_JAX}?{uuid_}"
    else:
        # installing wheel/sdist
        extras_require["jax"] = "sphericart-jax"

    setup(
        version=open(os.path.join("sphericart", "VERSION")).readline().strip(),
        ext_modules=[
            Extension(name="sphericart", sources=[]),
        ],
        cmdclass={
            "build_ext": cmake_ext,
            "bdist_egg": bdist_egg if "bdist_egg" in sys.argv else bdist_egg_disabled,
            "bdist_wheel": universal_wheel,
        },
        package_data={
            "sphericart": [
                "sphericart/lib/*",
                "sphericart/include/*",
            ],
        },
        extras_require=extras_require,
    )
